@@ -1,6 +1,6 @@
 /***************************************************************************
- *   Copyright (C) 2007 by PCMan (Hong Jen Yee)   *
- *   pcman.tw@gmail.com   *
+ *   Copyright (C) 2007 by PCMan (Hong Jen Yee) <pcman.tw@gmail.com>       *
+ *   Copyright (C) 2023 Ingo Brückl                                        *
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
  *   it under the terms of the GNU General Public License as published by  *
@@ -23,14 +23,29 @@
 
 static void image_view_finalize(GObject *iv);
 
-static void image_view_paint(  ImageView* iv, GdkEventExpose* evt );
 static void image_view_clear( ImageView* iv );
-static gboolean on_idle( ImageView* iv );
 static void calc_image_area( ImageView* iv );
+
+#if GTK_CHECK_VERSION(3, 0, 0)
+static void paint(  ImageView* iv, GdkRectangle* invalid_rect, GdkInterpType type, cairo_t* cr );
+
+static void image_view_paint(  ImageView* iv, cairo_t* cr );
+
+static void on_get_preferred_width( GtkWidget* widget, gint* minimal_width, gint* natural_width );
+static void on_get_preferred_height( GtkWidget* widget, gint* minimal_height, gint* natural_height );
+static gboolean on_draw_event(GtkWidget* widget, cairo_t* cr);
+
+#else // GTK2
+static gboolean on_idle( ImageView* iv );
 static void paint(  ImageView* iv, GdkRectangle* invalid_rect, GdkInterpType type );
 
+static void image_view_paint(  ImageView* iv, GdkEventExpose* evt );
+
 static void on_size_request( GtkWidget* w, GtkRequisition* req );
 static gboolean on_expose_event( GtkWidget* widget, GdkEventExpose* evt );
+#endif
+
+
 static void on_size_allocate( GtkWidget* widget, GtkAllocation    *allocation );
 
 G_DEFINE_TYPE( ImageView, image_view, GTK_TYPE_MISC );
@@ -60,9 +75,16 @@ void image_view_class_init( ImageViewClass* klass )
     obj_class->finalize = image_view_finalize;
 
     widget_class = GTK_WIDGET_CLASS ( klass );
+#if GTK_CHECK_VERSION(3, 0, 0)
+    widget_class->get_preferred_width = on_get_preferred_width;
+    widget_class->get_preferred_height = on_get_preferred_height;
+    widget_class->draw = on_draw_event;
+#else // GTK2
     widget_class->size_request = on_size_request;
-    widget_class->size_allocate = on_size_allocate;
     widget_class->expose_event = on_expose_event;
+#endif
+    widget_class->size_allocate = on_size_allocate;
+
 
 /*
     // set up scrolling support
@@ -133,6 +155,22 @@ void image_view_set_adjustments( ImageView* iv, GtkAdjustment* h, GtkAdjustment*
     }
 }
 
+#if GTK_CHECK_VERSION(3, 0, 0)
+
+void on_get_preferred_width( GtkWidget* widget, gint* minimal_width, gint* natural_width )
+{
+    ImageView* iv = (ImageView*)widget;
+    *minimal_width = *natural_width = iv->img_area.width;
+}
+
+void on_get_preferred_height( GtkWidget* widget, gint* minimal_height, gint* natural_height )
+{
+    ImageView* iv = (ImageView*)widget;
+    *minimal_height = *natural_height = iv->img_area.height;
+}
+
+#else // GTK2
+
 void on_size_request( GtkWidget* w, GtkRequisition* req )
 {
     ImageView* iv = (ImageView*)w;
@@ -143,10 +181,15 @@ void on_size_request( GtkWidget* w, GtkRequisition* req )
     GTK_WIDGET_CLASS(image_view_parent_class)->size_request (w, req);
 }
 
+#endif
+
 void on_size_allocate( GtkWidget* widget, GtkAllocation   *allocation )
 {
     GTK_WIDGET_CLASS(image_view_parent_class)->size_allocate( widget, allocation );
     ImageView* iv = (ImageView*)widget;
+
+    iv->allocation = *allocation;
+
 /*
     if( !iv->buffer || allocation->width != widget->allocation.width ||
         allocation->height != widget->allocation.height )
@@ -204,47 +247,35 @@ eel_cairo_get_clip_region (cairo_t *cr)
         return region;
 }
 
-gboolean on_draw_event( GtkWidget* widget, cairo_t *cr )
+static gboolean on_draw_event( GtkWidget* widget, cairo_t *cr )
 {
 
     ImageView* iv = (ImageView*)widget;
-    if ( gtk_widget_get_mapped(widget) ) 
+    if ( gtk_widget_get_mapped(widget) )
         image_view_paint( iv, cr );
     return FALSE;
 
 }
+
 void image_view_paint( ImageView* iv, cairo_t *cr )
 {
     if( iv->pix )
     {
-        GdkRectangle* rects = NULL;
-        int i = 0;
-        cairo_rectangle_int_t n_rects = 0;
-
-/*
-        int                 cairo_region_num_rectangles         (const cairo_region_t *region);
-        void                cairo_region_get_rectangle          (const cairo_region_t *region,
-                                                         int nth,
-                                                         cairo_rectangle_int_t *rectangle);
-
-void                gdk_region_get_rectangles           (const GdkRegion *region,
-                                                         GdkRectangle **rectangles,
-                                                         gint *n_rectangles);
-*/
-
-        cairo_region_get_rectangle(eel_cairo_get_clip_region(cr), &rects, &n_rects);
+        cairo_region_t * region = eel_cairo_get_clip_region(cr);
+        int n_rects = cairo_region_num_rectangles(region);
 
+        int i;
         for( i = 0; i < n_rects; ++i )
         {
-            paint( iv, rects + i, GDK_INTERP_NEAREST );
+            cairo_rectangle_int_t rectangle;
+            cairo_region_get_rectangle(region, i, &rectangle);
+            paint( iv, &rectangle, iv->interp_type, cr );
         }
-        g_free( rects );
 
-        if( 0 == iv->idle_handler )
-            iv->idle_handler = g_idle_add( (GSourceFunc)on_idle, iv );
+        cairo_region_destroy (region);
     }
 }
-#else
+#else // GTK2
 
 gboolean on_expose_event( GtkWidget* widget, GdkEventExpose* evt )
 {
@@ -275,66 +306,7 @@ void image_view_paint( ImageView* iv, GdkEventExpose* evt )
         for( i = 0; i < n_rects; ++i )
         {
             // GdkRectangle& rect = rects[i];
-            paint( iv, rects + i, GDK_INTERP_NEAREST );
-#if 0
-            g_debug("dirty dest: x=%d, y=%d, w=%d, h=%d\nx_off=%d, y_off=%d",
-                    rect.x, rect.y, rect.width, rect.height, iv->img_area.x, iv->img_area.y );
-
-            if( ! gdk_rectangle_intersect( &rect, &iv->img_area, &rect ) )
-                continue;
-
-            int dest_x = rect.x;
-            int dest_y = rect.y;
-
-            rect.x -= img_area.x;
-            rect.y -= img_area.y;
-
-            GdkPixbuf* src_pix = NULL;
-            int src_x, src_y;
-            if( iv->scale == 1.0 )  // original size
-            {
-                src_pix = (GdkPixbuf*)g_object_ref( pix );
-                src_x = rect.x;
-                src_y = rect.y;
-            }
-            else    // scaling is needed
-            {
-                GdkPixbuf* scaled_pix = NULL;
-                int src_w, src_h;
-                src_x = (int)floor( gdouble(rect.x) / scale + 0.5 );
-                src_y = (int)floor( gdouble(rect.y) / scale + 0.5 );
-                src_w = (int)floor( gdouble(rect.width) / scale + 0.5 );
-                src_h = (int)floor( gdouble(rect.height) / scale + 0.5 );
-                if( src_y > gdk_pixbuf_get_height(pix) )
-                    src_y = gdk_pixbuf_get_height(pix);
-                if( src_x + src_w > gdk_pixbuf_get_width(pix) )
-                    src_w = gdk_pixbuf_get_width(pix) - src_x;
-                if( src_y + src_h > gdk_pixbuf_get_height(pix) )
-                    src_h = gdk_pixbuf_get_height(pix) - src_y;
-                g_debug("orig src: x=%d, y=%d, w=%d, h=%d",
-                        src_x, src_y, src_w, src_h );
-
-                src_pix = gdk_pixbuf_new_subpixbuf( pix, src_x, src_y,  src_w, src_h );
-                scaled_pix = gdk_pixbuf_scale_simple( src_pix, rect.width, rect.height, interp_type );
-                g_object_unref( src_pix );
-                src_pix = scaled_pix;
-
-                src_x = 0;
-                src_y = 0;
-            }
-
-            if( G_LIKELY(src_pix) )
-            {
-                gdk_draw_pixbuf( widget->window,
-                                widget->style->fg_gc[GTK_STATE_NORMAL],
-                                src_pix,
-                                src_x, src_y,
-                                dest_x, dest_y,
-                                rect.width, rect.height,
-                                GDK_RGB_DITHER_NORMAL, 0, 0 );
-                g_object_unref( src_pix );
-            }
-#endif
+            paint( iv, rects + i, iv->interp_type );
         }
         g_free( rects );
 
@@ -385,10 +357,44 @@ void image_view_set_scale( ImageView* iv, gdouble new_scale, GdkInterpType type
     if( new_scale == iv->scale )
         return;
 //    gdouble factor = new_scale / iv->scale;
+
+    gint xPos, yPos;
+    gtk_widget_get_pointer(GTK_WIDGET(iv), &xPos, &yPos);
+
+    gdouble hadj_value = gtk_adjustment_get_value(iv->hadj);
+    gdouble vadj_value = gtk_adjustment_get_value(iv->vadj);
+
+    // if the pointer is not inside the window...
+    if (xPos < hadj_value || xPos > hadj_value + gtk_adjustment_get_page_size(iv->hadj) ||
+        yPos < vadj_value || yPos > vadj_value + gtk_adjustment_get_page_size(iv->vadj))
+    {
+        // ...pretend that it is over the center of the image
+        xPos = iv->img_area.width / 2;
+        yPos = iv->img_area.height / 2;
+    }
+
+    gdouble oldRelativePositionX =  (gdouble) xPos / iv->img_area.width;
+    gdouble oldRelativePositionY = (gdouble) yPos / iv->img_area.height;
+    gdouble visibleAreaX = xPos - hadj_value;
+    gdouble visibleAreaY = yPos - vadj_value;
+
     iv->scale = new_scale;
     if( iv->pix )
     {
         calc_image_area( iv );
+
+        hadj_value = oldRelativePositionX * iv->img_area.width - visibleAreaX;
+        vadj_value = oldRelativePositionY * iv->img_area.height - visibleAreaY;
+
+#if GTK_CHECK_VERSION(3, 0, 0)
+        gtk_adjustment_set_value(iv->hadj, hadj_value > 0 ? hadj_value : 0);
+        gtk_adjustment_set_value(iv->vadj, vadj_value > 0 ? vadj_value : 0);
+#else   // gtk_adjustment_set_value() triggers a "value-changed" event,
+        // causing image flickering in GTK+ 2 due to multiple paint calls
+        iv->hadj->value = hadj_value > 0 ? hadj_value : 0;
+        iv->vadj->value = vadj_value > 0 ? vadj_value : 0;
+#endif
+
         gtk_widget_queue_resize( (GtkWidget*)iv );
 //        gtk_widget_queue_draw( (GtkWidget*)iv );
 
@@ -416,6 +422,8 @@ void image_view_set_scale( ImageView* iv, gdouble new_scale, GdkInterpType type
     }
 }
 
+#if GTK_CHECK_VERSION(3, 0, 0)
+#else // GTK2
 gboolean on_idle( ImageView* iv )
 {
     GDK_THREADS_ENTER();
@@ -427,23 +435,31 @@ gboolean on_idle( ImageView* iv )
     if( G_LIKELY(iv->hadj) )
     {
         rect.x = (int)gtk_adjustment_get_value(iv->hadj);
+#if GTK_CHECK_VERSION(2, 14, 0)
+        rect.width = (int)gtk_adjustment_get_page_size(iv->hadj);
+#else
         rect.width = (int)iv->hadj->page_size;
+#endif
     }
     else
     {
-        rect.x = ((GtkWidget*)iv)->allocation.x;
-        rect.width = ((GtkWidget*)iv)->allocation.width;
+        rect.x = iv->allocation.x;
+        rect.width = iv->allocation.width;
     }
 
     if( G_LIKELY(iv->vadj) )
     {
         rect.y = (int)gtk_adjustment_get_value(iv->vadj);
+#if GTK_CHECK_VERSION(2, 14, 0)
+        rect.height = (int)gtk_adjustment_get_page_size(iv->vadj);
+#else
         rect.height = (int)iv->vadj->page_size;
+#endif
     }
     else
     {
-        rect.y = ((GtkWidget*)iv)->allocation.y;
-        rect.height = ((GtkWidget*)iv)->allocation.height;
+        rect.y = iv->allocation.y;
+        rect.height = iv->allocation.height;
     }
 
     paint( iv, &rect, iv->interp_type );
@@ -453,12 +469,13 @@ gboolean on_idle( ImageView* iv )
     iv->idle_handler = 0;
     return FALSE;
 }
+#endif
 
 void calc_image_area( ImageView* iv )
 {
     if( G_LIKELY( iv->pix ) )
     {
-        GtkAllocation allocation = ((GtkWidget*)iv)->allocation;
+        GtkAllocation allocation = iv->allocation;
         iv->img_area.width = (int)floor( ((gdouble)gdk_pixbuf_get_width( iv->pix )) * iv->scale + 0.5 );
         iv->img_area.height = (int)floor( ((gdouble)gdk_pixbuf_get_height( iv->pix )) * iv->scale + 0.5 );
         // g_debug( "width=%d, height=%d", width, height );
@@ -478,72 +495,62 @@ void calc_image_area( ImageView* iv )
     }
 }
 
+#if GTK_CHECK_VERSION(3, 0, 0)
+void paint( ImageView* iv, GdkRectangle* invalid_rect, GdkInterpType type, cairo_t* cr )
+#else // GTK2
 void paint( ImageView* iv, GdkRectangle* invalid_rect, GdkInterpType type )
+#endif
 {
     GdkRectangle rect;
     if( ! gdk_rectangle_intersect( invalid_rect, &iv->img_area, &rect ) )
         return;
 
-    int dest_x = rect.x;
-    int dest_y = rect.y;
+    int dest_x;
+    int dest_y;
+
+    GdkPixbuf* src_pix = NULL;
+    GdkPixbuf* scaled_pix = NULL;
+
+    dest_x = rect.x;
+    dest_y = rect.y;
 
     rect.x -= iv->img_area.x;
     rect.y -= iv->img_area.y;
 
-    GdkPixbuf* src_pix = NULL;
-    int src_x, src_y;
-    if( iv->scale == 1.0 )  // original size
-    {
-        src_pix = (GdkPixbuf*)g_object_ref( iv->pix );
-        src_x = rect.x;
-        src_y = rect.y;
-    }
-    else    // scaling is needed
-    {
-        GdkPixbuf* scaled_pix = NULL;
-        int src_w, src_h;
-        src_x = (int)floor( ((gdouble)rect.x) / iv->scale + 0.5 );
-        src_y = (int)floor( ((gdouble)rect.y) / iv->scale + 0.5 );
-        src_w = (int)floor( ((gdouble)rect.width) / iv->scale + 0.5 );
-        src_h = (int)floor( ((gdouble)rect.height) / iv->scale + 0.5 );
-        if( src_y > gdk_pixbuf_get_height( iv->pix ) )
-            src_y = gdk_pixbuf_get_height( iv->pix );
-        if( src_x + src_w > gdk_pixbuf_get_width( iv->pix ) )
-            src_w = gdk_pixbuf_get_width( iv->pix ) - src_x;
-        if( src_y + src_h > gdk_pixbuf_get_height( iv->pix ) )
-            src_h = gdk_pixbuf_get_height( iv->pix ) - src_y;
-        //g_debug("orig src: x=%d, y=%d, w=%d, h=%d",
-        //        src_x, src_y, src_w, src_h );
+    int src_x = (int)floor( ((gdouble)rect.x) / iv->scale + 0.5 );
+    int src_y = (int)floor( ((gdouble)rect.y) / iv->scale + 0.5 );
+    int src_w = (int)floor( ((gdouble)rect.width) / iv->scale + 0.5 );
+    int src_h = (int)floor( ((gdouble)rect.height) / iv->scale + 0.5 );
+    if( src_y > gdk_pixbuf_get_height( iv->pix ) )
+        src_y = gdk_pixbuf_get_height( iv->pix );
+    if( src_x + src_w > gdk_pixbuf_get_width( iv->pix ) )
+        src_w = gdk_pixbuf_get_width( iv->pix ) - src_x;
+    if( src_y + src_h > gdk_pixbuf_get_height( iv->pix ) )
+        src_h = gdk_pixbuf_get_height( iv->pix ) - src_y;
+    //g_debug("orig src: x=%d, y=%d, w=%d, h=%d",
+    //        src_x, src_y, src_w, src_h );
 
-        if ((src_w > 0) && (src_h > 0))
-        {
-            src_pix = gdk_pixbuf_new_subpixbuf( iv->pix, src_x, src_y,  src_w, src_h );
-            scaled_pix = gdk_pixbuf_scale_simple( src_pix, rect.width, rect.height, type );
-            g_object_unref( src_pix );
-            src_pix = scaled_pix;
-        }
-
-        src_x = 0;
-        src_y = 0;
+    if ((src_w > 0) && (src_h > 0))
+    {
+        src_pix = gdk_pixbuf_new_subpixbuf( iv->pix, src_x, src_y,  src_w, src_h );
+        scaled_pix = gdk_pixbuf_scale_simple( src_pix, rect.width, rect.height, type );
+        g_object_unref( src_pix );
+        src_pix = scaled_pix;
     }
 
     if( G_LIKELY(src_pix) )
     {
         GtkWidget* widget = (GtkWidget*)iv;
-/*
-        gdk_draw_pixbuf( widget->window,
-                         widget->style->fg_gc[GTK_STATE_NORMAL],
-                         src_pix,
-                         src_x, src_y,
-                         dest_x, dest_y,
-                         rect.width, rect.height,
-                         GDK_RGB_DITHER_NORMAL, 0, 0 );
-*/
-        // New function with cairo
+#if GTK_CHECK_VERSION(3, 0, 0)
+#else // GTK2
         cairo_t *cr = gdk_cairo_create (gtk_widget_get_window(widget));
+#endif
         gdk_cairo_set_source_pixbuf (cr, src_pix, dest_x, dest_y);
         cairo_paint (cr);
+#if GTK_CHECK_VERSION(3, 0, 0)
+#else // GTK2
         cairo_destroy (cr);
+#endif
 
         g_object_unref( src_pix );
     }
