@@ -1,6 +1,6 @@
 /***************************************************************************
- *   Copyright (C) 2007, 2008 by PCMan (Hong Jen Yee)                      *
- *   pcman.tw@gmail.com                                                    *
+ *   Copyright (C) 2007, 2008 by PCMan (Hong Jen Yee) <pcman.tw@gmail.com> *
+ *   Copyright (C) 2023 by Ingo Brückl                                     *
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
  *   it under the terms of the GNU General Public License as published by  *
@@ -27,6 +27,7 @@
 #include <glib/gi18n.h>
 #include <glib/gstdio.h>
 #include <gdk/gdkkeysyms.h>
+#include <gdk/gdkkeysyms-compat.h>
 
 #include <unistd.h>
 #include <string.h>
@@ -50,11 +51,18 @@ static GtkTargetEntry drop_targets[] =
     {"text/plain", 0, 1}
 };
 
+static gboolean show_toolbar;
+
+extern int ExifRotate(const char * fname, int new_angle);
+// defined in exif.c
+extern int ExifRotateFlipMapping[9][9];
+
 static void main_win_init( MainWin*mw );
 static void main_win_finalize( GObject* obj );
 
 static void create_nav_bar( MainWin* mw, GtkWidget* box);
 GtkWidget* add_nav_btn( MainWin* mw, const char* icon, const char* tip, GCallback cb, gboolean toggle);
+GtkWidget* add_nav_btn_img( MainWin* mw, const char* icon, const char* tip, GCallback cb, gboolean toggle, GtkWidget** ret_img);
 // GtkWidget* add_menu_item(  GtkMenuShell* menu, const char* label, const char* icon, GCallback cb, gboolean toggle=FALSE );
 static void rotate_image( MainWin* mw, int angle );
 static void show_popup_menu( MainWin* mw, GdkEventButton* evt );
@@ -63,6 +71,7 @@ static void show_popup_menu( MainWin* mw, GdkEventButton* evt );
 static gboolean on_delete_event( GtkWidget* widget, GdkEventAny* evt );
 static void on_size_allocate( GtkWidget* widget, GtkAllocation    *allocation );
 static gboolean on_win_state_event( GtkWidget* widget, GdkEventWindowState* state );
+static void on_scroll_size_allocate(GtkWidget* widget, GtkAllocation* allocation, MainWin* mv);
 static void on_zoom_fit( GtkToggleButton* btn, MainWin* mw );
 static void on_zoom_fit_menu( GtkMenuItem* item, MainWin* mw );
 static void on_full_screen( GtkWidget* btn, MainWin* mw );
@@ -75,10 +84,15 @@ static void on_rotate_clockwise( GtkWidget* btn, MainWin* mw );
 static void on_rotate_counterclockwise( GtkWidget* btn, MainWin* mw );
 static void on_save_as( GtkWidget* btn, MainWin* mw );
 static void on_save( GtkWidget* btn, MainWin* mw );
+static void cancel_slideshow(MainWin* mw);
+static gboolean next_slide(MainWin* mw);
+static void on_slideshow_menu( GtkMenuItem* item, MainWin* mw );
+static void on_slideshow( GtkToggleButton* btn, MainWin* mw );
 static void on_open( GtkWidget* btn, MainWin* mw );
 static void on_zoom_in( GtkWidget* btn, MainWin* mw );
 static void on_zoom_out( GtkWidget* btn, MainWin* mw );
 static void on_preference( GtkWidget* btn, MainWin* mw );
+static void on_toggle_toolbar( GtkMenuItem* item, MainWin* mw );
 static void on_quit( GtkWidget* btn, MainWin* mw );
 static gboolean on_button_press( GtkWidget* widget, GdkEventButton* evt, MainWin* mw );
 static gboolean on_button_release( GtkWidget* widget, GdkEventButton* evt, MainWin* mw );
@@ -90,6 +104,7 @@ static void on_drag_data_received( GtkWidget* widget, GdkDragContext *drag_conte
                 int x, int y, GtkSelectionData* data, guint info, guint time, MainWin* mw );
 static void on_delete( GtkWidget* btn, MainWin* mw );
 static void on_about( GtkWidget* menu, MainWin* mw );
+static gboolean on_animation_timeout( MainWin* mw );
 
 static void update_title(const char *filename, MainWin *mw );
 
@@ -98,6 +113,10 @@ void on_flip_horizontal( GtkWidget* btn, MainWin* mw );
 static int trans_angle_to_id(int i);
 static int get_new_angle( int orig_angle, int rotate_angle );
 
+static void main_win_set_zoom_scale(MainWin* mw, double scale);
+static void main_win_set_zoom_mode(MainWin* mw, ZoomMode mode);
+static void main_win_update_zoom_buttons_state(MainWin* mw);
+
 // Begin of GObject-related stuff
 
 G_DEFINE_TYPE( MainWin, main_win, GTK_TYPE_WINDOW )
@@ -127,8 +146,11 @@ void main_win_finalize( GObject* obj )
 
     if( G_LIKELY(mw->img_list) )
         image_list_free( mw->img_list );
+#if GTK_CHECK_VERSION(3, 0, 0)
+    g_object_unref( mw->hand_cursor );
+#else
     gdk_cursor_unref( mw->hand_cursor );
-
+#endif
     // FIXME: Put this here is weird
     gtk_main_quit();
 }
@@ -143,15 +165,31 @@ GtkWidget* main_win_new()
 void main_win_init( MainWin*mw )
 {
     gtk_window_set_title( (GtkWindow*)mw, _("Image Viewer"));
-    gtk_window_set_icon_from_file( (GtkWindow*)mw, PACKAGE_DATA_DIR"/pixmaps/gpicview.png", NULL );
-    gtk_window_set_default_size( (GtkWindow*)mw, 640, 480 );
+    if (gtk_icon_theme_has_icon(gtk_icon_theme_get_default(), "gpicview"))
+    {
+        gtk_window_set_icon_name((GtkWindow*)mw, "gpicview");
+    }
+    else
+    {
+        gtk_window_set_icon_from_file((GtkWindow*)mw, PACKAGE_DATA_DIR "/icons/hicolor/48x48/apps/gpicview.png", NULL);
+    }
+    gtk_window_move( GTK_WINDOW(mw), pref.win_x, pref.win_y );
+    gtk_window_set_default_size( GTK_WINDOW(mw), pref.win_w, pref.win_h );
 
+#if GTK_CHECK_VERSION(3, 0, 0)
+    GtkWidget* box = gtk_box_new( GTK_ORIENTATION_VERTICAL, 0 );
+#else
     GtkWidget* box = gtk_vbox_new( FALSE, 0 );
+#endif
     gtk_container_add( (GtkContainer*)mw, box);
 
     // image area
     mw->evt_box = gtk_event_box_new();
+#if GTK_CHECK_VERSION(2, 18, 0)
+    gtk_widget_set_can_focus(mw->evt_box,TRUE);
+#else
     GTK_WIDGET_SET_FLAGS( mw->evt_box, GTK_CAN_FOCUS );
+#endif
     gtk_widget_add_events( mw->evt_box,
                            GDK_POINTER_MOTION_MASK|GDK_BUTTON_PRESS_MASK|
                            GDK_BUTTON_RELEASE_MASK|GDK_SCROLL_MASK );
@@ -166,22 +204,34 @@ void main_win_init( MainWin*mw )
     mw->img_view = image_view_new();
     gtk_container_add( (GtkContainer*)mw->evt_box, (GtkWidget*)mw->img_view);
 
+#if GTK_CHECK_VERSION(3, 0, 0)
+#else
     const char scroll_style[]=
             "style \"gpicview-scroll\" {"
             "GtkScrolledWindow::scrollbar-spacing=0"
             "}"
             "class \"GtkScrolledWindow\" style \"gpicview-scroll\"";
     gtk_rc_parse_string( scroll_style );
+#endif
     mw->scroll = gtk_scrolled_window_new( NULL, NULL );
+    g_signal_connect(G_OBJECT(mw->scroll), "size-allocate", G_CALLBACK(on_scroll_size_allocate), (gpointer) mw);
     gtk_scrolled_window_set_shadow_type( (GtkScrolledWindow*)mw->scroll, GTK_SHADOW_NONE );
     gtk_scrolled_window_set_policy((GtkScrolledWindow*)mw->scroll,
                                     GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
     GtkAdjustment *hadj, *vadj;
     hadj = gtk_scrolled_window_get_hadjustment((GtkScrolledWindow*)mw->scroll);
+#if GTK_CHECK_VERSION(2, 14, 0)
+    gtk_adjustment_set_page_increment(hadj, 10);
+#else
     hadj->page_increment = 10;
+#endif
     gtk_adjustment_changed(hadj);
     vadj = gtk_scrolled_window_get_vadjustment((GtkScrolledWindow*)mw->scroll);
+#if GTK_CHECK_VERSION(2, 14, 0)
+    gtk_adjustment_set_page_increment(vadj, 10);
+#else
     vadj->page_increment = 10;
+#endif
     gtk_adjustment_changed(vadj);
 
     image_view_set_adjustments( IMAGE_VIEW(mw->img_view), hadj, vadj );    // dirty hack :-(
@@ -196,6 +246,14 @@ void main_win_init( MainWin*mw )
     create_nav_bar( mw, box );
     gtk_widget_show_all( box );
 
+    show_toolbar = pref.show_toolbar;
+
+    if (show_toolbar)
+        gtk_widget_show(gtk_widget_get_parent(mw->nav_bar));
+    else
+        gtk_widget_hide(gtk_widget_get_parent(mw->nav_bar));
+
+
     mw->hand_cursor = gdk_cursor_new_for_display( gtk_widget_get_display((GtkWidget*)mw), GDK_FLEUR );
 
 //    zoom_mode = ZOOM_NONE;
@@ -220,6 +278,7 @@ void create_nav_bar( MainWin* mw, GtkWidget* box )
 
     add_nav_btn( mw, GTK_STOCK_GO_BACK, _("Previous"), G_CALLBACK(on_prev), FALSE );
     add_nav_btn( mw, GTK_STOCK_GO_FORWARD, _("Next"), G_CALLBACK(on_next), FALSE );
+    mw->btn_play_stop = add_nav_btn_img( mw, GTK_STOCK_MEDIA_PLAY, _("Start Slideshow"), G_CALLBACK(on_slideshow), TRUE, &mw->img_play_stop );
 
     gtk_box_pack_start( (GtkBox*)mw->nav_bar, gtk_vseparator_new(), FALSE, FALSE, 0 );
 
@@ -268,6 +327,12 @@ void create_nav_bar( MainWin* mw, GtkWidget* box )
 
 gboolean on_delete_event( GtkWidget* widget, GdkEventAny* evt )
 {
+    if (pref.save_window)
+    {
+       gtk_window_get_position( GTK_WINDOW(widget), &pref.win_x, &pref.win_y );
+       gtk_window_get_size( GTK_WINDOW(widget), &pref.win_w, &pref.win_h );
+    }
+
     gtk_widget_destroy( widget );
     return TRUE;
 }
@@ -299,7 +364,9 @@ gboolean on_animation_timeout( MainWin* mw )
     int delay;
     if ( gdk_pixbuf_animation_iter_advance( mw->animation_iter, NULL ) )
     {
-        mw->pix = gdk_pixbuf_animation_iter_get_pixbuf( mw->animation_iter );
+        GdkPixbuf *current;
+        current = gdk_pixbuf_animation_iter_get_pixbuf( mw->animation_iter );
+        mw->pix = gdk_pixbuf_copy( current );
         image_view_set_pixbuf( (ImageView*)mw->img_view, mw->pix );
     }
     delay = gdk_pixbuf_animation_iter_get_delay_time( mw->animation_iter );
@@ -318,6 +385,16 @@ static void update_btns(MainWin* mw)
 
 gboolean main_win_open( MainWin* mw, const char* file_path, ZoomMode zoom )
 {
+    if (g_file_test(file_path, G_FILE_TEST_IS_DIR))
+    {
+        image_list_open_dir( mw->img_list, file_path, NULL );
+        image_list_sort_by_name( mw->img_list, GTK_SORT_DESCENDING );
+        if (image_list_get_first(mw->img_list))
+            main_win_open(mw, image_list_get_current_file_path(mw->img_list), zoom);
+        return TRUE;
+    }
+
+
     GError* err = NULL;
     GdkPixbufFormat* info;
     info = gdk_pixbuf_get_file_info( file_path, NULL, NULL );
@@ -375,19 +452,24 @@ gboolean main_win_open( MainWin* mw, const char* file_path, ZoomMode zoom )
         get_working_area( gtk_widget_get_screen((GtkWidget*)mw), &area );
         // g_debug("determine best zoom mode: orig size:  w=%d, h=%d", w, h);
         // FIXME: actually this is a little buggy :-(
-        if( w < area.width && h < area.height && (w >= 640 || h >= 480) )
+        if( w + pref.win_x < area.width && h + pref.win_y < area.height && (w >= pref.win_w || h >= pref.win_h) )
         {
             gtk_scrolled_window_set_policy( (GtkScrolledWindow*)mw->scroll, GTK_POLICY_NEVER, GTK_POLICY_NEVER );
             gtk_widget_set_size_request( (GtkWidget*)mw->img_view, w, h );
             GtkRequisition req;
             gtk_widget_size_request ( (GtkWidget*)mw, &req );
-            if( req.width < 640 )   req.width = 640;
-            if( req.height < 480 )   req.height = 480;
+            if( req.width < pref.win_w )   req.width = pref.win_w;
+            if( req.height < pref.win_h )   req.height = pref.win_h;
             gtk_window_resize( (GtkWindow*)mw, req.width, req.height );
             gtk_widget_set_size_request( (GtkWidget*)mw->img_view, -1, -1 );
             gtk_scrolled_window_set_policy( (GtkScrolledWindow*)mw->scroll, GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC );
             mw->zoom_mode = ZOOM_ORIG;
             mw->scale = 1.0;
+#if !GTK_CHECK_VERSION(3, 0, 0)
+            while (gtk_events_pending())
+                gtk_main_iteration();
+            g_usleep(10000);
+#endif
         }
         else
             mw->zoom_mode = ZOOM_FIT;
@@ -395,18 +477,14 @@ gboolean main_win_open( MainWin* mw, const char* file_path, ZoomMode zoom )
 
     if( mw->zoom_mode == ZOOM_FIT )
     {
-        gtk_toggle_button_set_active( (GtkToggleButton*)mw->btn_fit, TRUE );
         main_win_fit_window_size( mw, FALSE, GDK_INTERP_BILINEAR );
     }
     else  if( mw->zoom_mode == ZOOM_SCALE )  // scale
     {
-        gtk_toggle_button_set_active( (GtkToggleButton*)mw->btn_orig, FALSE );
-        gtk_toggle_button_set_active( (GtkToggleButton*)mw->btn_fit, FALSE );
         main_win_scale_image( mw, mw->scale, GDK_INTERP_BILINEAR );
     }
     else  if( mw->zoom_mode == ZOOM_ORIG )  // original size
     {
-        gtk_toggle_button_set_active( (GtkToggleButton*)mw->btn_orig, TRUE );
         image_view_set_scale( (ImageView*)mw->img_view, mw->scale, GDK_INTERP_BILINEAR );
         main_win_center_image( mw );
     }
@@ -431,16 +509,23 @@ gboolean main_win_open( MainWin* mw, const char* file_path, ZoomMode zoom )
     update_title( disp_name, mw );
     g_free( disp_name );
 
+    main_win_update_zoom_buttons_state(mw);
+
     return TRUE;
 }
 
+void main_win_start_slideshow( MainWin* mw )
+{
+    on_slideshow_menu(NULL, mw);
+}
+
 void main_win_close( MainWin* mw )
 {
     if( mw->animation )
     {
         g_object_unref( mw->animation );
         mw->animation = NULL;
-        if( mw->animation_timeout );
+        if( mw->animation_timeout )
         {
             g_source_remove( mw->animation_timeout );
             mw->animation_timeout = 0;
@@ -467,7 +552,11 @@ void main_win_show_error( MainWin* mw, const char* message )
 void on_size_allocate( GtkWidget* widget, GtkAllocation    *allocation )
 {
     GTK_WIDGET_CLASS(main_win_parent_class)->size_allocate( widget, allocation );
+#if GTK_CHECK_VERSION(2, 20, 0)
+    if(gtk_widget_get_realized (widget) )
+#else
     if( GTK_WIDGET_REALIZED (widget) )
+#endif
     {
         MainWin* mw = (MainWin*)widget;
 
@@ -484,7 +573,7 @@ void on_size_allocate( GtkWidget* widget, GtkAllocation    *allocation )
 gboolean on_win_state_event( GtkWidget* widget, GdkEventWindowState* state )
 {
     MainWin* mw = (MainWin*)widget;
-    if( state->new_window_state == GDK_WINDOW_STATE_FULLSCREEN )
+    if( (state->new_window_state & GDK_WINDOW_STATE_FULLSCREEN) != 0 )
     {
         gtk_widget_modify_bg( mw->evt_box, GTK_STATE_NORMAL, &pref.bg_full );
         gtk_widget_hide( gtk_widget_get_parent(mw->nav_bar) );
@@ -493,14 +582,11 @@ gboolean on_win_state_event( GtkWidget* widget, GdkEventWindowState* state )
     else
     {
         gtk_widget_modify_bg( mw->evt_box, GTK_STATE_NORMAL, &pref.bg );
-        gtk_widget_show( gtk_widget_get_parent(mw->nav_bar) );
+        if (show_toolbar)
+            gtk_widget_show( gtk_widget_get_parent(mw->nav_bar) );
         mw->full_screen = FALSE;
     }
 
-    int previous = pref.open_maximized;
-    pref.open_maximized = (state->new_window_state == GDK_WINDOW_STATE_MAXIMIZED);
-    if (previous != pref.open_maximized)
-        save_preferences();
     return TRUE;
 }
 
@@ -535,11 +621,16 @@ void main_win_fit_window_size(  MainWin* mw, gboolean can_strech, GdkInterpType
 
     if( mw->pix == NULL )
         return;
-
-    main_win_fit_size( mw, mw->scroll->allocation.width, mw->scroll->allocation.height, can_strech, type );
+    main_win_fit_size( mw, mw->scroll_allocation.width, mw->scroll_allocation.height, can_strech, type );
 }
 
-GtkWidget* add_nav_btn( MainWin* mw, const char* icon, const char* tip, GCallback cb, gboolean toggle )
+GtkWidget* add_nav_btn( MainWin* mw, const char* icon, const char* tip, GCallback cb, gboolean toggle)
+{
+    GtkWidget* unused;
+    return add_nav_btn_img(mw, icon, tip, cb, toggle, &unused);
+}
+
+GtkWidget* add_nav_btn_img( MainWin* mw, const char* icon, const char* tip, GCallback cb, gboolean toggle, GtkWidget** ret_img )
 {
     GtkWidget* img;
     if( g_str_has_prefix(icon, "gtk-") )
@@ -562,9 +653,15 @@ GtkWidget* add_nav_btn( MainWin* mw, const char* icon, const char* tip, GCallbac
     gtk_container_add( (GtkContainer*)btn, img );
     gtk_widget_set_tooltip_text( btn, tip );
     gtk_box_pack_start( (GtkBox*)mw->nav_bar, btn, FALSE, FALSE, 0 );
+    *ret_img = img;
     return btn;
 }
 
+void on_scroll_size_allocate(GtkWidget* widget, GtkAllocation* allocation, MainWin* mv)
+{
+    mv->scroll_allocation = *allocation;
+}
+
 void on_zoom_fit_menu( GtkMenuItem* item, MainWin* mw )
 {
     gtk_button_clicked( (GtkButton*)mw->btn_fit );
@@ -572,15 +669,8 @@ void on_zoom_fit_menu( GtkMenuItem* item, MainWin* mw )
 
 void on_zoom_fit( GtkToggleButton* btn, MainWin* mw )
 {
-    if( ! btn->active )
-    {
-        if( mw->zoom_mode == ZOOM_FIT )
-            gtk_toggle_button_set_active( btn, TRUE );
-        return;
-    }
-    gtk_toggle_button_set_active( (GtkToggleButton*)mw->btn_orig, FALSE );
-
-    main_win_fit_window_size( mw, FALSE, GDK_INTERP_BILINEAR );
+    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(btn)))
+        main_win_set_zoom_mode(mw, ZOOM_FIT);
 }
 
 void on_full_screen( GtkWidget* btn, MainWin* mw )
@@ -605,31 +695,8 @@ void on_orig_size( GtkToggleButton* btn, MainWin* mw )
         return;
     }
 
-    if( ! btn->active )
-    {
-        if( mw->zoom_mode == ZOOM_ORIG )
-            gtk_toggle_button_set_active( btn, TRUE );
-        return;
-    }
-    mw->zoom_mode = ZOOM_ORIG;
-    mw->scale = 1.0;
-//    gtk_scrolled_window_set_policy( (GtkScrolledWindow*)mw->scroll,
-//                                     GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC );
-
-    gtk_toggle_button_set_active( (GtkToggleButton*)mw->btn_fit, FALSE );
-
-    if( ! mw->pix )
-        return;
-
-    // update scale
-    update_title(NULL, mw);
-
-    image_view_set_scale( (ImageView*)mw->img_view, 1.0, GDK_INTERP_BILINEAR );
-
-    while (gtk_events_pending ())
-        gtk_main_iteration ();
-
-    main_win_center_image( mw ); // FIXME:  mw doesn't work well. Why?
+    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(btn)))
+        main_win_set_zoom_mode(mw, ZOOM_ORIG);
 }
 
 void on_prev( GtkWidget* btn, MainWin* mw )
@@ -649,7 +716,7 @@ void on_prev( GtkWidget* btn, MainWin* mw )
     if( name )
     {
         char* file_path = image_list_get_current_file_path( mw->img_list );
-        main_win_open( mw, file_path, ZOOM_FIT );
+        main_win_open( mw, file_path, mw->zoom_mode );
         g_free( file_path );
     }
 }
@@ -670,11 +737,54 @@ void on_next( GtkWidget* btn, MainWin* mw )
     if( name )
     {
         char* file_path = image_list_get_current_file_path( mw->img_list );
-        main_win_open( mw, file_path, ZOOM_FIT );
+        main_win_open( mw, file_path, mw->zoom_mode );
         g_free( file_path );
     }
 }
 
+void cancel_slideshow(MainWin* mw)
+{
+    mw->slideshow_cancelled = TRUE;
+    mw->slideshow_running = FALSE;
+    if (mw->slide_timeout != 0)
+        g_source_remove(mw->slide_timeout);
+    gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(mw->btn_play_stop), FALSE );
+}
+
+gboolean next_slide(MainWin* mw)
+{
+    /* Timeout causes an implicit "Next". */
+    if (mw->slideshow_running)
+        on_next( NULL, mw );
+
+    return mw->slideshow_running;
+}
+
+void on_slideshow_menu( GtkMenuItem* item, MainWin* mw )
+{
+    gtk_button_clicked( (GtkButton*)mw->btn_play_stop );
+}
+
+void on_slideshow( GtkToggleButton* btn, MainWin* mw )
+{
+    if ((mw->slideshow_running) || (mw->slideshow_cancelled))
+    {
+        mw->slideshow_running = FALSE;
+        mw->slideshow_cancelled = FALSE;
+        gtk_image_set_from_stock( GTK_IMAGE(mw->img_play_stop), GTK_STOCK_MEDIA_PLAY, GTK_ICON_SIZE_SMALL_TOOLBAR );
+        gtk_widget_set_tooltip_text( GTK_WIDGET(btn), _("Start Slideshow") );
+        gtk_toggle_button_set_active( btn, FALSE );
+    }
+    else
+    {
+        gtk_toggle_button_set_active( btn, TRUE );
+        mw->slideshow_running = TRUE;
+        gtk_image_set_from_stock( GTK_IMAGE(mw->img_play_stop), GTK_STOCK_MEDIA_STOP, GTK_ICON_SIZE_SMALL_TOOLBAR );
+        gtk_widget_set_tooltip_text( GTK_WIDGET(btn), _("Stop Slideshow") );
+        mw->slide_timeout = g_timeout_add(1000 * pref.slide_delay, (GSourceFunc) next_slide, mw);
+    }
+}
+
 //////////////////// rotate & flip
 
 static int trans_angle_to_id(int i)
@@ -686,13 +796,12 @@ static int trans_angle_to_id(int i)
     else if(i == -45)	return 7;
     else if(i == -90)	return 2;
     else if(i == -135)	return 5;
-    else if(i == -180)	return 4;
+    else     /* -180 */ return 4;
 }
 
 static int get_new_angle( int orig_angle, int rotate_angle )
 {
     // defined in exif.c
-    extern int ExifRotateFlipMapping[9][9];
     static int angle_trans_back[] = {0, 0, -90, 180, -180, -135, 90, -45, 270};
 
     orig_angle = trans_angle_to_id(orig_angle);
@@ -713,6 +822,7 @@ void on_rotate_auto_save( GtkWidget* btn, MainWin* mw )
 
 void on_rotate_clockwise( GtkWidget* btn, MainWin* mw )
 {
+    cancel_slideshow(mw);
     rotate_image( mw, GDK_PIXBUF_ROTATE_CLOCKWISE );
     mw->rotation_angle = get_new_angle(mw->rotation_angle, 90);
     on_rotate_auto_save(btn, mw);
@@ -720,6 +830,7 @@ void on_rotate_clockwise( GtkWidget* btn, MainWin* mw )
 
 void on_rotate_counterclockwise( GtkWidget* btn, MainWin* mw )
 {
+    cancel_slideshow(mw);
     rotate_image( mw, GDK_PIXBUF_ROTATE_COUNTERCLOCKWISE );
     mw->rotation_angle = get_new_angle(mw->rotation_angle, 270);
     on_rotate_auto_save(btn, mw);
@@ -727,6 +838,7 @@ void on_rotate_counterclockwise( GtkWidget* btn, MainWin* mw )
 
 void on_flip_vertical( GtkWidget* btn, MainWin* mw )
 {
+    cancel_slideshow(mw);
     rotate_image( mw, -180 );
     mw->rotation_angle = get_new_angle(mw->rotation_angle, -180);
     on_rotate_auto_save(btn, mw);
@@ -734,6 +846,7 @@ void on_flip_vertical( GtkWidget* btn, MainWin* mw )
 
 void on_flip_horizontal( GtkWidget* btn, MainWin* mw )
 {
+    cancel_slideshow(mw);
     rotate_image( mw, -90 );
     mw->rotation_angle = get_new_angle(mw->rotation_angle, -90);
     on_rotate_auto_save(btn, mw);
@@ -745,6 +858,7 @@ void on_save_as( GtkWidget* btn, MainWin* mw )
 {
     char *file, *type;
 
+    cancel_slideshow(mw);
     if( ! mw->pix )
         return;
 
@@ -776,6 +890,7 @@ void on_save_as( GtkWidget* btn, MainWin* mw )
 
 void on_save( GtkWidget* btn, MainWin* mw )
 {
+    cancel_slideshow(mw);
     if( ! mw->pix )
         return;
 
@@ -819,6 +934,7 @@ void on_save( GtkWidget* btn, MainWin* mw )
 
 void on_open( GtkWidget* btn, MainWin* mw )
 {
+    cancel_slideshow(mw);
     char* file = get_open_filename( (GtkWindow*)mw, image_list_get_dir( mw->img_list ) );
     if( file )
     {
@@ -829,47 +945,16 @@ void on_open( GtkWidget* btn, MainWin* mw )
 
 void on_zoom_in( GtkWidget* btn, MainWin* mw )
 {
-    mw->zoom_mode = ZOOM_SCALE;
-    gtk_toggle_button_set_active( (GtkToggleButton*)mw->btn_fit, FALSE );
-    gtk_toggle_button_set_active( (GtkToggleButton*)mw->btn_orig, FALSE );
-//    gtk_scrolled_window_set_policy( (GtkScrolledWindow*)mw->scroll,
-//                                     GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC );
-
     double scale = mw->scale;
-    if( mw->pix && scale < 20.0 )
-    {
-//        busy(TRUE);
-            scale *= 1.05;
-            if( scale > 20.0 )
-                scale = 20.0;
-            if( mw->scale != scale )
-                main_win_scale_image( mw, scale, GDK_INTERP_BILINEAR );
-//        adjust_adjustment_on_zoom(oldscale);
-//        busy(FALSE);
-    }
+    scale *= 1 + pref.zoom_factor / 100.0;
+    main_win_set_zoom_scale(mw, scale);
 }
 
 void on_zoom_out( GtkWidget* btn, MainWin* mw )
 {
-    mw->zoom_mode = ZOOM_SCALE;
-    gtk_toggle_button_set_active( (GtkToggleButton*)mw->btn_fit, FALSE );
-    gtk_toggle_button_set_active( (GtkToggleButton*)mw->btn_orig, FALSE );
-//    gtk_scrolled_window_set_policy( (GtkScrolledWindow*)mw->scroll,
-//                                     GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC );
-
     double scale = mw->scale;
-    if( mw->pix && scale > 0.02 )
-    {
-//        busy(TRUE);
-
-        scale /= 1.05;
-        if( scale < 0.02 )
-            scale = 0.02;
-        if( mw->scale != scale )
-            main_win_scale_image( mw, scale, GDK_INTERP_BILINEAR );
-//        adjust_adjustment_on_zoom(oldscale);
-//        busy(FALSE);
-    }
+    scale /= 1 + pref.zoom_factor / 100.0;
+    main_win_set_zoom_scale(mw, scale);
 }
 
 void on_preference( GtkWidget* btn, MainWin* mw )
@@ -879,12 +964,17 @@ void on_preference( GtkWidget* btn, MainWin* mw )
 
 void on_quit( GtkWidget* btn, MainWin* mw )
 {
+    cancel_slideshow(mw);
     gtk_widget_destroy( (GtkWidget*)mw );
 }
 
 gboolean on_button_press( GtkWidget* widget, GdkEventButton* evt, MainWin* mw )
 {
+#if GTK_CHECK_VERSION(2, 14, 0)
+    if( ! gtk_widget_has_focus( widget ) )
+#else
     if( ! GTK_WIDGET_HAS_FOCUS( widget ) )
+#endif
         gtk_widget_grab_focus( widget );
 
     if( evt->type == GDK_BUTTON_PRESS)
@@ -895,7 +985,7 @@ gboolean on_button_press( GtkWidget* widget, GdkEventButton* evt, MainWin* mw )
                 return FALSE;
             mw->dragging = TRUE;
             gtk_widget_get_pointer( (GtkWidget*)mw, &mw->drag_old_x ,&mw->drag_old_y );
-            gdk_window_set_cursor( widget->window, mw->hand_cursor );
+            gdk_window_set_cursor( gtk_widget_get_window(widget), mw->hand_cursor );
         }
         else if( evt->button == 3 )   // right button
         {
@@ -927,35 +1017,57 @@ gboolean on_mouse_move( GtkWidget* widget, GdkEventMotion* evt, MainWin* mw )
     GtkRequisition req;
     gtk_widget_size_request( (GtkWidget*)mw->img_view, &req );
 
+#if GTK_CHECK_VERSION(2, 14, 0)
+    gdouble hadj_page_size = gtk_adjustment_get_page_size(hadj);
+    gdouble hadj_lower = gtk_adjustment_get_lower(hadj);
+    gdouble hadj_upper = gtk_adjustment_get_upper(hadj);
+#else
+    gdouble hadj_page_size = hadj->page_size;
+    gdouble hadj_lower = hadj->lower;
+    gdouble hadj_upper = hadj->upper;
+#endif
+
     if( ABS(dx) > 4 )
     {
         mw->drag_old_x = cur_x;
-        if( req.width > hadj->page_size )
+        if( req.width > hadj_page_size )
         {
-            gdouble x = gtk_adjustment_get_value (hadj) + dx;
-            if( x < hadj->lower )
-                x = hadj->lower;
-            else if( (x + hadj->page_size) > hadj->upper )
-                x = hadj->upper - hadj->page_size;
+            gdouble value = gtk_adjustment_get_value (hadj);
+            gdouble x = value + dx;
+            if( x < hadj_lower )
+                x = hadj_lower;
+            else if( (x + hadj_page_size) > hadj_upper )
+                x = hadj_upper - hadj_page_size;
 
-            if( x != hadj->value )
+            if( x != value )
                 gtk_adjustment_set_value (hadj, x );
         }
     }
 
+#if GTK_CHECK_VERSION(2, 14, 0)
+    gdouble vadj_page_size = gtk_adjustment_get_page_size(vadj);
+    gdouble vadj_lower = gtk_adjustment_get_lower(vadj);
+    gdouble vadj_upper = gtk_adjustment_get_upper(vadj);
+#else
+    gdouble vadj_page_size = vadj->page_size;
+    gdouble vadj_lower = vadj->lower;
+    gdouble vadj_upper = vadj->upper;
+#endif
+
     if( ABS(dy) > 4 )
     {
-        if( req.height > vadj->page_size )
+        if( req.height > vadj_page_size )
         {
             mw->drag_old_y = cur_y;
-            gdouble y = gtk_adjustment_get_value (vadj) + dy;
-            if( y < vadj->lower )
-                y = vadj->lower;
-            else if( (y + vadj->page_size) > vadj->upper )
-                y = vadj->upper - vadj->page_size;
+            gdouble value = gtk_adjustment_get_value (vadj);
+            gdouble y = value + dy;
+            if( y < vadj_lower )
+                y = vadj_lower;
+            else if( (y + vadj_page_size) > vadj_upper )
+                y = vadj_upper - vadj_page_size;
 
-            if( y != vadj->value )
-                gtk_adjustment_set_value (vadj, y  );
+            if( y != value )
+                gtk_adjustment_set_value (vadj, y );
         }
     }
     return FALSE;
@@ -964,7 +1076,7 @@ gboolean on_mouse_move( GtkWidget* widget, GdkEventMotion* evt, MainWin* mw )
 gboolean on_button_release( GtkWidget* widget, GdkEventButton* evt, MainWin* mw )
 {
     mw->dragging = FALSE;
-    gdk_window_set_cursor( widget->window, NULL );
+    gdk_window_set_cursor( gtk_widget_get_window(widget), NULL );
     return FALSE;
 }
 
@@ -1037,6 +1149,10 @@ gboolean on_key_press_event(GtkWidget* widget, GdkEventKey * key)
         case GDK_uparrow:
             on_prev( NULL, mw );
             break;
+        case GDK_w:
+        case GDK_W:
+            on_slideshow_menu( NULL, mw );
+            break;
         case GDK_KP_Add:
         case GDK_plus:
         case GDK_equal:
@@ -1093,6 +1209,10 @@ gboolean on_key_press_event(GtkWidget* widget, GdkEventKey * key)
         case GDK_P:
             on_preference( NULL, mw );
 	    break;
+        case GDK_t:
+        case GDK_T:
+            on_toggle_toolbar( NULL, mw );
+	    break;
         case GDK_Escape:
             if( mw->full_screen )
                 on_full_screen( NULL, mw );
@@ -1122,11 +1242,27 @@ void main_win_center_image( MainWin* mw )
     GtkRequisition req;
     gtk_widget_size_request( (GtkWidget*)mw->img_view, &req );
 
-    if( req.width > hadj->page_size )
-        gtk_adjustment_set_value(hadj, ( hadj->upper - hadj->page_size ) / 2 );
+#if GTK_CHECK_VERSION(2, 14, 0)
+    gdouble hadj_page_size = gtk_adjustment_get_page_size(hadj);
+    gdouble hadj_upper = gtk_adjustment_get_upper(hadj);
+#else
+    gdouble hadj_page_size = hadj->page_size;
+    gdouble hadj_upper = hadj->upper;
+#endif
 
-    if( req.height > vadj->page_size )
-        gtk_adjustment_set_value(vadj, ( vadj->upper - vadj->page_size ) / 2 );
+    if( req.width > hadj_page_size )
+        gtk_adjustment_set_value(hadj, ( hadj_upper - hadj_page_size ) / 2 );
+
+#if GTK_CHECK_VERSION(2, 14, 0)
+    gdouble vadj_page_size = gtk_adjustment_get_page_size(vadj);
+    gdouble vadj_upper = gtk_adjustment_get_upper(vadj);
+#else
+    gdouble vadj_page_size = vadj->page_size;
+    gdouble vadj_upper = vadj->upper;
+#endif
+
+    if( req.height > vadj_page_size )
+        gtk_adjustment_set_value(vadj, ( vadj_upper - vadj_page_size ) / 2 );
 }
 
 void rotate_image( MainWin* mw, int angle )
@@ -1253,6 +1389,7 @@ gboolean main_win_save( MainWin* mw, const char* file_path, const char* type, gb
 
 void on_delete( GtkWidget* btn, MainWin* mw )
 {
+    cancel_slideshow(mw);
     char* file_path = image_list_get_current_file_path( mw->img_list );
     if( file_path )
     {
@@ -1302,12 +1439,24 @@ void on_delete( GtkWidget* btn, MainWin* mw )
     }
 }
 
+void on_toggle_toolbar( GtkMenuItem* item, MainWin* mw )
+{
+    show_toolbar = !show_toolbar;
+
+    if (show_toolbar)
+        gtk_widget_show(gtk_widget_get_parent(mw->nav_bar));
+    else
+        gtk_widget_hide(gtk_widget_get_parent(mw->nav_bar));
+}
+
+
 void show_popup_menu( MainWin* mw, GdkEventButton* evt )
 {
     static PtkMenuItemEntry menu_def[] =
     {
         PTK_IMG_MENU_ITEM( N_( "Previous" ), GTK_STOCK_GO_BACK, on_prev, GDK_leftarrow, 0 ),
         PTK_IMG_MENU_ITEM( N_( "Next" ), GTK_STOCK_GO_FORWARD, on_next, GDK_rightarrow, 0 ),
+        PTK_IMG_MENU_ITEM( N_( "Start/Stop Slideshow" ), GTK_STOCK_MEDIA_PLAY, on_slideshow_menu, GDK_W, 0 ),
         PTK_SEPARATOR_MENU_ITEM,
         PTK_IMG_MENU_ITEM( N_( "Zoom Out" ), GTK_STOCK_ZOOM_OUT, on_zoom_out, GDK_minus, 0 ),
         PTK_IMG_MENU_ITEM( N_( "Zoom In" ), GTK_STOCK_ZOOM_IN, on_zoom_in, GDK_plus, 0 ),
@@ -1328,6 +1477,7 @@ void show_popup_menu( MainWin* mw, GdkEventButton* evt )
         PTK_IMG_MENU_ITEM( N_("Delete File"), GTK_STOCK_DELETE, G_CALLBACK(on_delete), GDK_Delete, 0 ),
         PTK_SEPARATOR_MENU_ITEM,
         PTK_IMG_MENU_ITEM( N_("Preferences"), GTK_STOCK_PREFERENCES, G_CALLBACK(on_preference), GDK_P, 0 ),
+        PTK_IMG_MENU_ITEM( N_("Show/Hide toolbar"), NULL, G_CALLBACK(on_toggle_toolbar), GDK_T, 0 ),
         PTK_STOCK_MENU_ITEM( GTK_STOCK_ABOUT, on_about ),
         PTK_SEPARATOR_MENU_ITEM,
         PTK_IMG_MENU_ITEM( N_("Quit"), GTK_STOCK_QUIT, G_CALLBACK(on_quit), GDK_Q, 0 ),
@@ -1360,28 +1510,6 @@ void show_popup_menu( MainWin* mw, GdkEventButton* evt )
     gtk_menu_popup( (GtkMenu*)popup, NULL, NULL, NULL, NULL, evt->button, evt->time );
 }
 
-/* callback used to open default browser when URLs got clicked */
-static void open_url( GtkAboutDialog *dlg, const gchar *url, gpointer data)
-{
-    /* FIXME: is there any better way to do this? */
-    char* programs[] = { "xdg-open", "gnome-open", "exo-open" };
-    int i;
-    for(  i = 0; i < G_N_ELEMENTS(programs); ++i)
-    {
-        gchar* open_cmd = NULL;
-        if( (open_cmd = g_find_program_in_path( programs[i] )) )
-        {
-             char* argv [3];
-             argv [0] = programs[i];
-             argv [1] = (gchar *) url;
-             argv [2] = NULL;
-             g_spawn_async (NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, NULL);
-             g_free( open_cmd );
-             break;
-        }
-    }
-}
-
 void on_about( GtkWidget* menu, MainWin* mw )
 {
     GtkWidget * about_dlg;
@@ -1392,21 +1520,20 @@ void on_about( GtkWidget* menu, MainWin* mw )
         "Hialan Liu <hialan.liu@gmail.com>",
         "Marty Jack <martyj19@comcast.net>",
         "Louis Casillas <oxaric@gmail.com>",
+        "Will Davies",
         _(" * Refer to source code of EOG image viewer and GThumb"),
         NULL
     };
     /* TRANSLATORS: Replace this string with your names, one name per line. */
     gchar *translators = _( "translator-credits" );
 
-    gtk_about_dialog_set_url_hook( open_url, mw, NULL);
-
     about_dlg = gtk_about_dialog_new ();
 
     gtk_container_set_border_width ( ( GtkContainer*)about_dlg , 2 );
     gtk_about_dialog_set_version ( (GtkAboutDialog*)about_dlg, VERSION );
-    gtk_about_dialog_set_name ( (GtkAboutDialog*)about_dlg, _( "GPicView" ) );
-    gtk_about_dialog_set_logo( (GtkAboutDialog*)about_dlg, gdk_pixbuf_new_from_file(  PACKAGE_DATA_DIR"/pixmaps/gpicview.png", NULL ) );
-    gtk_about_dialog_set_copyright ( (GtkAboutDialog*)about_dlg, _( "Copyright (C) 2007 - 2009" ) );
+    gtk_about_dialog_set_program_name ( (GtkAboutDialog*)about_dlg, _( "GPicView" ) );
+    gtk_about_dialog_set_logo_icon_name ( (GtkAboutDialog*)about_dlg, "gpicview" );
+    gtk_about_dialog_set_copyright ( (GtkAboutDialog*)about_dlg, _( "Copyright (C) 2007 - 2023" ) );
     gtk_about_dialog_set_comments ( (GtkAboutDialog*)about_dlg, _( "Lightweight image viewer from LXDE project" ) );
     gtk_about_dialog_set_license ( (GtkAboutDialog*)about_dlg, "GPicView\n\nCopyright (C) 2007 Hong Jen Yee (PCMan)\n\nThis program is free software; you can redistribute it and/or\nmodify it under the terms of the GNU General Public License\nas published by the Free Software Foundation; either version 2\nof the License, or (at your option) any later version.\n\nThis program is distributed in the hope that it will be useful,\nbut WITHOUT ANY WARRANTY; without even the implied warranty of\nMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\nGNU General Public License for more details.\n\nYou should have received a copy of the GNU General Public License\nalong with this program; if not, write to the Free Software\nFoundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA." );
     gtk_about_dialog_set_website ( (GtkAboutDialog*)about_dlg, "http://wiki.lxde.org/en/GPicView" );
@@ -1421,7 +1548,16 @@ void on_about( GtkWidget* menu, MainWin* mw )
 void on_drag_data_received( GtkWidget* widget, GdkDragContext *drag_context,
                 int x, int y, GtkSelectionData* data, guint info, guint time, MainWin* mw )
 {
-    if( ! data || data->length <= 0)
+    if( !data)
+        return;
+
+#if GTK_CHECK_VERSION(2, 14, 0)
+    int data_length = gtk_selection_data_get_length(data);
+#else
+    int data_length = data->length;
+#endif
+
+    if( data_length <= 0)
         return;
 
     char* file = NULL;
@@ -1444,3 +1580,60 @@ void on_drag_data_received( GtkWidget* widget, GdkDragContext *drag_context,
         g_free( file );
     }
 }
+
+static void main_win_set_zoom_scale(MainWin* mw, double scale)
+{
+    main_win_set_zoom_mode(mw, ZOOM_SCALE);
+
+    if (scale > 20.0)
+        scale = 20.0;
+    if (scale < 0.02)
+        scale = 0.02;
+
+    if (mw->scale != scale)
+        main_win_scale_image(mw, scale, GDK_INTERP_BILINEAR);
+}
+
+static void main_win_set_zoom_mode(MainWin* mw, ZoomMode mode)
+{
+    if (mw->zoom_mode == mode)
+       return;
+
+    mw->zoom_mode = mode;
+
+    main_win_update_zoom_buttons_state(mw);
+
+    if (mode == ZOOM_ORIG)
+    {
+        mw->scale = 1.0;
+        if (!mw->pix)
+           return;
+
+        update_title(NULL, mw);
+
+        image_view_set_scale( (ImageView*)mw->img_view, 1.0, GDK_INTERP_BILINEAR );
+
+        while (gtk_events_pending ())
+            gtk_main_iteration ();
+
+        main_win_center_image( mw ); // FIXME:  mw doesn't work well. Why?
+    }
+    else if (mode == ZOOM_FIT)
+    {
+        main_win_fit_window_size( mw, FALSE, GDK_INTERP_BILINEAR );
+    }
+}
+
+static void main_win_update_zoom_buttons_state(MainWin* mw)
+{
+    gboolean button_fit_active = mw->zoom_mode == ZOOM_FIT;
+    gboolean button_orig_active = mw->zoom_mode == ZOOM_ORIG;
+
+    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mw->btn_fit)) != button_fit_active)
+        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mw->btn_fit), button_fit_active);
+
+    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mw->btn_orig)) != button_orig_active)
+        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mw->btn_orig), button_orig_active);
+}
+
+
